/* -*- Mode: c++ -*- */
/*
 * Copyright 2001 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/*
 *  Copyright 1997 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */

#ifndef _VRAUDIOSOURCE_H_
#define _VRAUDIOSOURCE_H_

#define AUDIOINCHUNKSIZE 200

extern "C" {
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
}

#include <VrSource.h>
#include <fstream>
#include <algorithm>
#include <string>

template<class oType>
class VrAudioSource : public VrSource<oType> {
protected:
  int         audiofd;
  long        format;
  bool	      decimate;
  std::string device_name;
public: 
  virtual int work2(VrSampleRange output, void *o[]);
  virtual void initialize();

  VrAudioSource(double sampling_freq, const char* dev = "/dev/dsp")
    : audiofd(-1), decimate(false), device_name(dev){

    setSamplingFrequency (sampling_freq);
    setOutputSize (AUDIOINCHUNKSIZE);
  }

  virtual const char *name() { return "VrAudioSource"; }

  virtual ~VrAudioSource() {close(audiofd);}
};


template<class oType> void
VrAudioSource<oType>::initialize() 
{
  int temp = 0x7fff0004;
  if(audiofd==-1) {
    if ((audiofd = open(device_name.c_str(), O_RDONLY)) < 0) {
      cerr << "VrAudioSource: ";
      perror (device_name.c_str ());
      exit(1);
    }
    if ((ioctl(audiofd,SNDCTL_DSP_SETFRAGMENT,&temp)) < 0) {
      fprintf (stderr, "VrAudioSource: set fragment returned %d\n", errno);
      exit(1);
    }
  }

  if(sizeof(oType)==1) {
    format=AFMT_MU_LAW;
  } else if(sizeof(oType)==2) {
    printf("setting to 16 bits \n");
    format = AFMT_S16_LE;
  } else {
    fprintf (stderr,
	     "VrAudioSource: cannot handle data type of size %d\n",
	     sizeof(oType));
    exit(1);
  }

  ioctl(audiofd,SNDCTL_DSP_RESET);

  int origf=format;
  if((ioctl(audiofd,SNDCTL_DSP_SETFMT,&format)) < 0) {
    cerr << "VrAudioSource: " << device_name << " IOCTL failed with errno " << errno << "\n";
    exit(1);
  }
  if(origf!=format) {
    fprintf (stderr, "VrAudioSource: Warning: unable to support format %d\n", origf);
    fprintf (stderr, "  card requested %ld instead.\n", format);
  }

  // set number of channels no matter what.  Some hardware only does stereo

  int noutputs = 1;	// number of outputs
  int channels = noutputs;
  if (ioctl (audiofd, SNDCTL_DSP_CHANNELS, &channels) < 0){
    perror ("VrAudioSink: SNDCTL_DSP_CHANNELS failed");
    exit (1);
  }
    
  if (channels != noutputs){
    if (noutputs == 2){
      fprintf(stderr, "VrAudioSource: could not set STEREO mode\n");
      // FIXME, could fix this by duplicating mono input, or zeroing other channel
      exit(1);
    }
    else {
      // user asked for MONO, but got STEREO.  We can work around this...
      fprintf (stderr, "VrAudioSource: could not set MONO mode\n");
      decimate = true;
    }
  }

  int sf = (int)getSamplingFrequency();
  //cout <<"Sampling frequency = "<<sf<<endl;
  if ((ioctl(audiofd,SNDCTL_DSP_SPEED,&sf)) < 0) {
    cerr << device_name << ": Invalid sampling frequency...defaulting to 8 Khz\n";
    sf = 8000;
    if ((ioctl(audiofd,SNDCTL_DSP_SPEED,&sf)) < 0) {
      fprintf (stderr, "Couldn't even manage that...aborting\n");
      exit(1);
    }
  }
  if(sf!=getSamplingFrequency())
    fprintf(stderr, "VrAudioSource Warning: soundcard defaulted to %d Hz\n", sf);
}


template<class oType> int 
VrAudioSource<oType>::work2(VrSampleRange output, void *ao[])
{
  unsigned units=output.size;
  oType **o= (oType **)ao;
  oType *curPos=o[0];

  if (decimate){
    // convert STEREO input from sound card to MONO output
    while(units > 0) {
      const unsigned N = 2048;
      oType	tmp[N*2];
      
      int nbytes =  std::min (units, N) * (sizeof(oType) * 2);
      int count = read(audiofd, tmp, nbytes);

      if(count < 0) {
	perror ("VrAudioSource");
	exit(1);
      }

      assert ((count & ((sizeof(oType) * 2) - 1)) == 0);

      for (unsigned int i = 0; i < count / (sizeof (oType) * 2); i++){
	curPos[i] = tmp[2*i];
      }

      units -= count / (sizeof (oType) * 2);
      curPos += count / (sizeof (oType) * 2);
    }
    return output.size;
  }

  else {	// normal
    while(units > 0) {
      int nbytes =  units * sizeof(oType);
      int count = read(audiofd, curPos, nbytes);

      if(count < 0) {
	perror ("VrAudioSource");
	exit(1);
      }

      assert ((count & (sizeof(oType) - 1)) == 0);

      units -= count / sizeof (oType);
      curPos += count / sizeof (oType);
    }
    return output.size;
  }
}

#endif
